#!/usr/bin/perl

# spell.cgi - a Web Service returning lists of alternative spelling; perfect for me  ;-)

# Eric Lease Morgan <emorgan@nd.edu>
# September 21, 2005 - Added DTD; call it version 0.1
# September 12, 2005 - First cut; stress reliever

# POD

=head1 NAME

spell.cgi


=head1 DESCRIPTION

This CGI script implements a simple Web Service-based called spell. Given a URL that looks like this:

  http://spell.example.org/?word=origami

the script will return an XML stream looking like this:

  <spell>
      <version>0.1</version>
      <word>origami</word>
      <dictionary>master</dictionary>
      <spellings>
          <spelling>origami</spelling>
          <spelling>origem</spelling>
          <spelling>irrigam</spelling>
          <spelling>obrigam</spelling>
      </spellings>
  </spell>

User-agents can then take the XML and provide value-added services for users, such as alternative spelling to queries returning few or no hits.


=head1 INSTALLATION

=over 4

=item 1. 

Install the GNU dictionary program called Aspell, as well as its Perl interface.

=item 2. 

Create at least one Aspell dictionary. For example, all the words from one of your indexes.

=item 3.  

Create a tab-delimited file (say, dictionaries.db) where the first item denotes the name of the dictionary, and the second item denotes the full-path of your newley created Aspell dictionary.

=item 4.  

Copy spell.cgi to a location of your choice, as well as the etc directory containing your tab-delimited file.

=item 5.  

Define the script's constants, specifically: 1) the full-path to your tab-delimited file (DICTIONARIES), 2) the default dictionary to query (DICTIONARY), and 3) the location of the default dictionary (PATH). You should not have to define VERSION unless you improve the spell XML output.

=item 6. 

Use your favorite user-agent to test your work.

=back

By creating more than one dictionary it is possible to refine the spell service. One dictionary might be strong on science. Another might be strong on history. Yet a third might be an amalgamation of the first two -- a general dictionary. Update the contents of the tab-delimited file and clients will be able to query selected dictionaries using an additional name-value pair in the spell URL:

  http://spell.example.org/?word=FOO&dictionary=BAR
  
FOO is the query word, and BAR is the dictionary to search.


=head1 AUTHORS

Eric Lease Morgan <emorgan@nd.edu> & Team Ockham (September 21, 2005)

=cut


# define constants
use constant DICTIONARIES => './etc/dictionaries.db';
use constant DICTIONARY   => 'master';
use constant PATH         => '/home/ND/mylibrary/htdocs/DATA/master_dict';
use constant VERSION      => '0.1';


######################################################
# no configuration should be necessary below this line

# require
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use strict;
use Text::Aspell;

# initialize
my $cgi        = CGI->new;
my $dictionary = DICTIONARY;
my $path       = PATH;
my $header     = &header;
my $version    = VERSION;
my $response;

# get the input
my $word = $cgi->param('word');

# existential check
if (! $word ) {

	$response .= &error( 'No word' );
	&gracefulExit;

}

# slurp up the dictionary database and find the necessary path
my %dictionaries = &read_dictionaries();
if ($cgi->param('dictionary')) {
	foreach my $record ( values(%dictionaries) ) {
	
		if ( $record->{'key'} eq $cgi->param('dictionary') ) {
		
			$dictionary = $record->{'key'};
			$path       = $record->{'path'};
		
		}
		
	}

}

# spell check and a bit of packaging
my $spellings = &spellings2xml(&spell(( word => $word, dictionary => $path)));

# build the response
$response .= <<RESPONSE;
$header
<spell>
	<version>$version</version>
	<word>$word</word>
	<dictionary>$dictionary</dictionary>
	<spellings>
		$spellings
	</spellings>
</spell>
RESPONSE

# quit
&gracefulExit;


sub read_dictionaries {
	
	# read a tab-delimited file defined by DICTIONARIES and return it as a hash
	my %db = {};
	open ( INPUT, '< ' . DICTIONARIES );
	while ( <INPUT> ) {
	
		chop;
		my ($key, $path) = split /\t/;
		my $index               = {};
		$index->{'key'}         = $key;
		$index->{'path'}        = $path;
		$db{$key}               = $index;
		
	}
	
	close INPUT;
	return %db;
	
}


sub spell {

	# get the input
	my %args = @_;
		
	# the reason de existance
	my $speller = Text::Aspell->new;
	$speller->set_option('master', $args{'dictionary'} );
	return $speller->suggest( $args{'word'} );

}


sub spellings2xml {

	# wrap stuff in a simple element
	my @spellings = @_;
	my $spellings;
	foreach (@spellings) { $spellings .= "<spelling>" . $_ . '</spelling>' . "\n\t\t" }
	return $spellings;

}


sub gracefulExit {

	print $cgi->header(-type=>'text/xml', -charset => 'UTF-8');
	print $response;
	exit;

}


sub error {

	my $msg = $_[0];
	
	return <<XML
$header
<spell>
	<version>$version</version>
	<error>$msg</error>
</spell>
XML

}


sub header {

	return <<HEADER
<?xml version='1.0' encoding='UTF-8'?>
<!DOCTYPE spell
[
 <!ELEMENT spell      (version, ((error) | (word, dictionary, spellings)))>
 <!ELEMENT version    (#PCDATA)>
 <!ELEMENT error      (#PCDATA)>
 <!ELEMENT word       (#PCDATA)>
 <!ELEMENT dictionary (#PCDATA)>
 <!ELEMENT spellings  (spelling+)>
 <!ELEMENT spelling   (#PCDATA)>
]>
<!-- 

  For more information, see:
  
    http://spell.ockham.org/about/
    
  This XML file, defined by the included DTD is the result of simple Web 
  Service called spell.                                                  
	                                                                  
  The purpose of the Service is to return alternative spellings to given 
  words. Queries to the Service take the following form:                 
	                                   
    http://spell.ockham.org/?word=FOO&dictionary=BAR 
  
  where FOO is a word to look up and BAR can be any one of the 
  following values: 
  
    * biological 
    * earth 
    * library 
    * master (the default) 
    * mathematics 
  
  Here are a number of sample queries: 
  
    http://spell.ockham.org/?word=origami 
    http://spell.ockham.org/?word=origami&dictionary=master 
    http://spell.ockham.org/?word=origami&dictionary=earth 
    http://spell.ockham.org/?word=m0rgan&dictionary=library 
    
  Author: Eric Lease Morgan 
  Email: emorgan\@nd.edu    
  Date created: September 20, 2005
  Date updated: September 21, 2005

-->

HEADER

}
